"""Top-level package for lognflow."""

__author__ = 'Alireza Sadri'
__email__ = 'arsadri@gmail.com'
__version__ = '0.12.11'

from .lognflow import lognflow
from .logviewer import logviewer
from .printprogress import printprogress
from .plt_utils import (
    plt_colorbar, plot_gaussian_gradient, plt_imshow, plt_violinplot,
    plt_imhist, transform3D_viewer)
from .utils import (
    select_directory, select_file, repr_raw, replace_all, 
    is_builtin_collection, text_to_collection, stack_to_frame, 
    stacks_to_frames, ssh_system)
from .multiprocessor import multiprocessor
from .loopprocessor import loopprocessor

getLogger = lognflow
def basicConfig(*args, **kwargs):
    ...